# -*- coding: utf-8 -*-
"""Functions from market data"""

__author__ = "Miguel Martin"
__version__ = "1"

import numpy as np

# Get maximum drawdown of Serie
def max_drawdown_value(vec):
    maximums = np.maximum.accumulate(vec)
    drawdowns = maximums - vec
    return np.max(drawdowns)


def max_drawdown_ratio(vec):
    maximums = np.maximum.accumulate(vec)
    drawdowns = 1 - vec / maximums
    return np.max(drawdowns)


# Safe sum. If list is empty return Nan
def safe_sum(l):
    if len(l) == 0:
        return np.NaN
    return np.sum(l)


# Safe div. If list is empty return Nan
def safe_div(a, b):
    if b == 0:
        return np.NaN
    return a / b


# Safe min. If list is empty return Nan
def safe_min(l):
    if len(l) == 0:
        return 0
    return -np.min(l)


def safe_mean(l):
    if len(l) == 0:
        return np.NaN
    return np.mean(l)

def default_comission_cost(num_acciones, precio_compra, precio_venta):
    return 0

def default_slippage_cost(num_acciones, precio_compra, precio_venta):
    return 0


class Policy:
    """
    Represents the mathematical rationale of a quantitative investing strategy

    Attributes
    ----------
    long_stop_loss : float
        Stop loss when the position is long
    short_stop_loss : float
        Stop loss when the position is short
    long_stop_loss_trailling : float
        Trailling stop loss when the position is long
    short_stop_loss_trailling : float
        Trailling stop loss when the position is short

    Methods
    -------
    execute(cls, day, data, trades)
        Returns the investing actions to take
    sell_short_when()
        Returns the function to analyse of a short position should be opened
    buy_long_when()
        Returns the function to analyse if a long position should be opened
    close_short_when()
        Returns the functions to analyse if a short position should be closed
    close_long_when()
        Returns the functions to analyse if a long position should be closed
    """

    plot_functions = []
    second_plot_functions = []

    long_stop_loss = 1
    short_stop_loss = 1
    long_stop_loss_trailling = False
    short_stop_loss_trailling = False

    ticker = "Hello"  # This remains so that the system does not break

    @staticmethod
    def sell_short_when():
        return lambda day, ticker, trades, data: False

    @staticmethod
    def buy_long_when():
        return lambda day, ticker, trades, data: False

    @staticmethod
    def close_short_when():
        return lambda day, ticker, trades, data: False

    @staticmethod
    def close_long_when():
        return lambda day, ticker, trades, data: False

    @classmethod
    def execute(cls, day, data, trades):
        actions = []

        f = cls.close_long_when()
        if f(day, cls.ticker, trades, data):
            actions += ["Close_long"]

        f = cls.close_short_when()
        if f(day, cls.ticker, trades, data):
            actions += ["Close_short"]

        f = cls.sell_short_when()
        if f(day, cls.ticker, trades, data):
            actions += ["Sell_short"]

        f = cls.buy_long_when()
        if f(day, cls.ticker, trades, data):
            actions += ["Buy_long"]

        return actions


def build_policy(policy_dict):
    created_policy = type(policy_dict["name"], (Policy,), {})
    for attribute in policy_dict["policy"]["parameters"]:
        setattr(
            created_policy, attribute, policy_dict["policy"]["parameters"][attribute]
        )
    for function in policy_dict["policy"]["functions"]:
        function_definition = policy_dict["policy"]["functions"][function]
        setattr(
            created_policy, function, lambda: set_head_function(function_definition)
        )
    return created_policy


def set_head_function(function_definition):
    module_name = next(iter(function_definition))
    function_name = function_definition[module_name]
    params_definition = function_definition["params"]
    params = []
    head_function = getattr(globals()[module_name], function_name)
    for param in params_definition:
        if next(iter(param)) in ["functions", "indicators"]:
            params.append(set_head_function(param))
        else:
            params.append(next(iter(param.values())))
    return head_function(*params)



# -*- coding: utf-8 -*-
"""
Functions to create strategies.

All of the functions return a *programatic function* which calculates the given function for some OCHL data.

To build a function not included in the package, use the following template::

    def function_name(params, target='close'):
        def return_function(data):
            your code goes here
        return return_function

"""

import numpy as np
import pandas as pd

def line(value):
    """
    | Line which can be used to cross with functions like RSI or MACD.
    | Name: line\_\ **value**\ 

    :param value: Value of the line
    :type value: float
    """
    def return_function(data):
        column_name = f'line_{value}'
        if column_name not in data.columns:
            data[column_name] = value
        return data[column_name].copy()

    return return_function

def days_to_constant(days, order=1):
    #TODO: Finish documenting
    """
    .. _days_to_constant:

    Calculates the constant for exponential smoothing from a given window size

    :param days: Window size
    :type days: int
    :param order: [description], defaults to 1
    :type order: int, optional
    """
    return float("%.3f" % (1 - pow(1 - 2 / (days + 1), 1 / order)))


def get_column(column):
    def return_function(data):
        return data[column].copy()

    return return_function

def trailing(target="close", days=40, over_under="under"):
    """
    | Calculates the if the target is trailing under or over the current in the past days.
    | Name: trailling\_\ **over\_under**\ \_\ **days**\ \_of\_\ **target**

    :param target: Data column to use for the calculation, defaults to "close"
    :type target: str
    :param days: Size of the window in days, defaults to 40
    :type days: int
    """
    def f():
        def g(x):
            return all(x[-1] > x[:-1])

        return g

    def return_function(data):
        column_name = f"trailling_{over_under}_{days}_of_{target}"
        if column_name not in data.columns:
            data[column_name] = (
                data[target].rolling(window=days, min_periods=1).apply(f(), raw=True)
            )
        return data[column_name].copy()

    return return_function


def rolling_std(days, target="close"):
    """
    | Calculates the rolling standard deviation
    | Name: rolling\_std\_\ **days**\ \_of\_\ **target**

    :param days: Size of the window in days
    :type days: int
    :param target: Data column to use for the calculation, defaults to "close"
    :type target: str, optional
    """
    def return_function(data):
        column_name = f"rolling_std_{days}_of_{target}"
        if column_name not in data.columns:
            data[column_name] = data[target].rolling(days, min_periods=2).std()
        return data[column_name].copy()

    return return_function


def moving_average(days, target="close"):
    """
    | Calculates the rolling moving average
    | Name: moving\_average\_\ **days**\ \_of\_\ **target**

    :param days: Size of the window in days
    :type days: int
    :param target: Data column to use for the calculation, defaults to "close"
    :type target: str, optional
    """
    def return_function(data):
        column_name = f"moving_average_{days}_of_{target}"
        if column_name not in data.columns:
            data[column_name] = data[target].rolling(days, min_periods=1).mean()
        return data[column_name].copy()

    return return_function


def weighted_moving_average(days, weights):
    """
    | Calculates the rolling weighted moving average. The weights list should have the size of the days window.
    | Name: weighted\_moving\_average\_\ **days**

    :param days: Size of the window
    :type days: int
    :param weights: Weights to use for the calculation
    :type weights: list(int)
    """
    def f(w):
        def g(x):
            return (w * x).sum() / sum(w)

        return g

    def return_function(data):
        if len(weights) == days:
            column_name = f"weighted_moving_average_{days}"
            if column_name not in data.columns:
                data[column_name] = (
                    data["close"].rolling(window=days).apply(f(weights), raw=True)
                )
            return data[column_name].copy()
        else:
            raise Exception("Weights length and rolling mean must be same length")

    return return_function


def step_weighting_ma(days, first_weight=1, step=1):
    """
    | Calculates the weighted moving average using stepped weights
    | Name: step\_weighting\_ma\_\ **days**

    :param days: Window size
    :type days: int
    :param first_weight: Weight of the first data point, defaults to 1
    :type first_weight: int, optional
    :param step: Weight increment for each data point, defaults to 1
    :type step: int, optional
    """
    def f(w):
        def g(x):
            return (w * x).sum() / sum(w)

        return g

    def return_function(data):
        weights = np.linspace(first_weight, (days * step) + first_weight - step, days)
        column_name = f"step_weighting_ma_{days}"
        if column_name not in data.columns:
            data[column_name] = (
                data["close"].rolling(window=days).apply(f(weights), raw=True)
            )
        return data[column_name].copy()

    return return_function


def percentage_weighting_ma(days, last_weight=1, step=0.5):
    """
    | Calculates the weighted moving average using backwards set weights. F.e. for a given last days=3, weight=1, step=0.5, the wieghts would be [0.25, 0.5, 1]
    | Name: percentage\_weighting\_ma\_\ **days**

    :param days: Window size
    :type days: int
    :param last_weight: Weights of the last data point, defaults to 1
    :type last_weight: int, optional
    :param step: Weight multiplier for each previous data point, defaults to 0.5
    :type step: float, optional
    """
    def f(w):
        def g(x):
            return (w * x).sum() / sum(w)

        return g

    def return_function(data):
        weights = np.zeros(days)
        weights[-1] = last_weight
        for i in range(1, days):
            weights[-1 * (i + 1)] = weights[-1 * i] * step
        column_name = f"percentage_weighting_ma_{days}"
        if column_name not in data.columns:
            data[column_name] = (
                data["close"].rolling(window=days).apply(f(weights), raw=True)
            )
        return data[column_name].copy()

    return return_function


def triangular_weighting_ma(days=10, shape="linear"):
    """
    | Calculates the weighted moving average with the weights having a triangular shape
    | Name: **shape**\ \_trianguar\_weighting\_ma\_\ **days**

    :param days: Window size, defaults to 10
    :type days: int, optional
    :param shape: Type of shape. Can be "linear" or "gaussian" , defaults to "linear"
    :type shape: str, optional
    """
    def f(w):
        def g(x):
            return (w * x).sum() / sum(w)

        return g

    def linear_triangular_weights():
        weights = np.zeros(days)
        weights[0 : int((days + 2) / 2) - 1] = range(1, int((days + 2) / 2))
        if days % 2 == 0:
            weights[int((days + 2) / 2) - 1 :] = range(1, int((days + 2) / 2))[::-1]
        else:
            weights[int((days + 2) / 2) - 1 :] = range(1, int((days + 2) / 2) + 1)[::-1]
        return weights

    def gaussian_triangular_weights():
        return NotImplementedError()

    if shape not in ["linear", "gaussian"]:
        raise NotImplementedError()
    else:
        weight_options = {
            "linear": linear_triangular_weights,
            "gaussian": gaussian_triangular_weights,
        }
        weights_func = weight_options.get(shape)
        weights_values = weights_func()

    def return_function(data):
        column_name = f"{shape}_trianguar_weighting_ma_{days}"
        if column_name not in data.columns:
            data[column_name] = (
                data["close"].rolling(window=days).apply(f(weights_values), raw=True)
            )
        return data[column_name].copy()

    return return_function


def pivot_point_weighting_ma(days):
    #TODO: Review the implementation of this function. Pivot point in the internet is referred by the internet to the moving average
    #of (close+high+low)/3. Refer to the book.
    def f(w):
        def g(x):
            return (w * x).sum() / sum(w)

        return g

    def pivot_point_weights():
        weights = np.array(range(1, days + 1))
        weights = weights * 3 - days - 1
        return weights

    def return_function(data):
        weights = pivot_point_weights()
        column_name = f"pivot_point_weighting_ma_{days}"
        if column_name not in data.columns:
            data[column_name] = (
                data["close"].rolling(window=days).apply(f(weights), raw=True)
            )
        return data[column_name].copy()

    return return_function


def geometric_moving_average(days, target='close'):
    """
    | Calculates the geometric moving average
    | Name: geometric\_moving\_average\_\ **days**

    :param days: Window size
    :type days: int
    :param target: Data column to use, defaults to "close"
    :type target: str, optional
    """
    def return_function(data):
        column_name = f"geometric_moving_average_{days}"
        if column_name not in data.columns:
            data[column_name] = pow(
                data[target].rolling(window=days).apply(np.prod, raw=True), 1 / days
            )
        return data[column_name].copy()

    return return_function


def exponential_smoothing(constant, target="close"):
    """
    | Calculates the exponential smoothing
    | Name: exponential\_smoothing\_\ **constant**\ %\_of\_\ **target**

    :param constant: Smoothing constant for the calculation. Use function days_to_constant_
    :type constant: float
    :param target: Data column to use, defaults to "close"
    :type target: str, optional
    """
    def return_function(data):
        column_name = f"exponential_smoothing_{constant}%_of_{target}"
        if column_name not in data.columns:
            data[column_name] = data[target].copy()
            for i in data[data[target].notnull()].index[1:]:
                data.loc[i, column_name] = data[column_name].loc[:i][-2] + constant * (
                    data[target].loc[i] - data[column_name].loc[:i][-2])

        return data[column_name].copy()

    return return_function


def second_order_exponential_smoothing(constant):
    """
    | Calculates the second order exponential smoothing of close
    | Name: second\_order\_exponential\_smoothing\_\ **constant**\ %

    :param constant: Smoothing constant for the calculation. Use function days_to_constant_
    :type constant: float
    """
    def return_function(data):
        first_order = exponential_smoothing(constant)(data)
        column_name = f"second_order_exponential_smoothing_{constant}%"
        if column_name not in data.columns:
            data[column_name] = first_order.copy()
            for i in range(1, len(data["close"])):
                data[column_name].iloc[i] = data[column_name].iloc[i - 1] + constant * (
                    first_order.iloc[i] - data[column_name].iloc[i - 1]
                )
        return data[column_name].copy()

    return return_function


def third_order_exponential_smoothing(constant):
    """
    | Calculates the second order exponential smoothing of close
    | Name: third\_order\_exponential\_smoothing\_\ **constant**\ %

    :param constant: Smoothing constant for the calculation. Use function days_to_constant_
    :type constant: float
    """
    def return_function(data):
        second_order = second_order_exponential_smoothing(constant)(data)
        column_name = f"third_order_exponential_smoothing_{constant}%"
        if column_name not in data.columns:
            data[column_name] = second_order.copy()
            for i in range(1, len(data["close"])):
                data[column_name].iloc[i] = data[column_name].iloc[i - 1] + constant * (
                    second_order.iloc[i] - data[column_name].iloc[i - 1]
                )
        return data[column_name].copy()

    return return_function


def lag_correction_exponential_smoothing(constant):
    #TODO: View the book for explanation
    def return_function(data):
        column_name = (f"lag_correction_exponential_smoothing_{constant*100}%")
        if column_name not in data.columns:
            first_order = exponential_smoothing(constant)(data)
            data[column_name] = data["close"] - first_order
            for i in range(1, len(data["close"])):
                data[column_name].iloc[i] = data[column_name].iloc[i - 1] + constant * (
                    data[column_name].iloc[i] - data[column_name].iloc[i - 1]
                )
            data[column_name] = data[column_name] + first_order
        return data[column_name].copy()

    return return_function


def double_moving_average(days):
    """
    | Calculates the double moving average of close
    | Name: double\_moving\_average\_\ **days**

    :param days: Window size
    :type days: int
    """
    def return_function(data):
        column_name = f"double_moving_average_{days}"
        if column_name not in data.columns:
            simple_ma = moving_average(days)(data)
            data[column_name] = simple_ma.rolling(days, min_periods=1).mean()
        return data[column_name].copy()

    return return_function


def double_smoothed_momentum(days1, days2):
    """
    | Calculates the momentum, smoothed by two consecutive exponential smoothings
    | Name: DSM\ **days1**\ \/\ **days2**
    
    :param days1: First smoothing window size
    :type days1: int
    :param days2: Second smoothing window size
    :type days2: int
    """
    def return_function(data):
        column_name = f"DSM{days1}/{days2}"
        if column_name not in data.columns:
            momentum_val = momentum(days=days1)(data)
            constant = days_to_constant(days1)
            exp_smoothed = exponential_smoothing(constant, target=momentum_val.name)(
                data
            )
            constant = days_to_constant(days2, 2)
            data[column_name] = exponential_smoothing(
                constant, target=exp_smoothed.name
            )(data)
        return data[column_name].copy()

    return return_function


def regularized_exponential_ma(days, weight, target="close"):
    #TODO: Refer to book
    def return_function(data):
        column_name = f"REMA_{days}/{weight}_of_{target}"
        if column_name not in data.columns:
            c = days_to_constant(days)
            data[column_name] = data[target].copy()
            for i in range(2, len(data)):
                data[column_name].iloc[i] = (
                    data[column_name].iloc[i - 1] * (1 + 2 * weight)
                    + c * (data[column_name].iloc[i] - data[column_name].iloc[i - 1])
                    - weight * data[column_name].iloc[i - 2]
                ) / (1 + weight)
        return data[column_name].copy()

    return return_function


def hull_moving_average(period=16, target="close"):
    """
    | Calculates the hull moving average
    | Name: hull\_moving\_average\_\ **period**\ \_of\_\ **target**

    :param period: Window size, defaults to 16
    :type period: int, optional
    :param target: Data colum to use, defaults to "close"
    :type target: str, optional
    """
    def return_function(data):
        column_name = f"hull_moving_average_{period}_of_{target}"
        if column_name not in data.columns:
            period_ma = moving_average(period)(data)
            period2_ma = moving_average(int(period / 2))(data)
            data[column_name] = (
                (2 * period2_ma - period_ma)
                .rolling(int(pow(period, 1 / 2)), min_periods=1)
                .mean()
            )
        return data[column_name].copy()

    return return_function


def upper_keltner_channel(days=10):
    """
    | Calculates the upper keltner channel
    | Name: upper\_keltner\_\ **days**

    :param days: Window size, defaults to 10
    :type days: int, optional
    """
    def return_function(data):
        column_name = f"upper_keltner_{days}"
        if column_name not in data.columns:
            average_daily_price = (data["close"] + data["high"] + data["low"]) / 3
            if (f"moving_average_{days}_of_close") not in data.columns:
                moving_average(days)(data)
            data[column_name] = (
                data[f"moving_average_{days}_of_close"] + average_daily_price
            )
        return data[column_name].copy()

    return return_function


def lower_keltner_channel(days=10):
    """
    | Calculates the lower keltner channel
    | Name: lower\_keltner\_\ **days**

    :param days: Window size, defaults to 10
    :type days: int, optional
    """
    def return_function(data):
        column_name = f"lower_keltner_{days}"
        if column_name not in data.columns:
            average_daily_price = (data["close"] + data["high"] + data["low"]) / 3
            if (f"moving_average_{days}_of_close") not in data.columns:
                moving_average(days)(data)
            data[column_name] = (
                data[f"moving_average_{days}_of_close"] - average_daily_price
            )
        return data[column_name].copy()

    return return_function


def upper_percentage_band(c, band_target, center_target):
    """
    | Calculates the upper percentage band
    | Name: upper\_percentage\_band\_\ **c**\ \_of\_\ **band\_target.name**\ \_over\_\ **center_target.name**

    :param c: multiplier constant
    :type c: float
    :param band_target: Data column used for the band displacement
    :type band_target: str
    :param center_target: Data column used for the center of the channel
    :type center_target: str
    """
    def return_function(data):
        column_name = f"upper_percentage_band_{c}_of_{band_target.name}_over_{center_target.name}"
        if column_name not in data.columns:
            print(center_target)
            data[column_name] = c * band_target + center_target.values
        return data[column_name].copy()

    return return_function


def lower_percentage_band(c, band_target, center_target):
    """
    | Calculates the lower percentage band
    | Name: lower\_percentage\_band\_\ **c**\ \_of\_\ **band\_target.name**\ \_over\_\ **center_target.name**

    :param c: multiplier constant
    :type c: float
    :param band_target: Data column used for the band displacement
    :type band_target: str
    :param center_target: Data column used for the center of the channel
    :type center_target: str
    """
    def return_function(data):
        column_name = f"lower_percentage_band_{c}_of_{band_target.name}_under_{center_target.name}"
        if column_name not in data.columns:
            data[column_name] = -c * band_target + center_target.values
        return data[column_name].copy()

    return return_function


def upper_absolute_band(value, target):
    """
    | Calculates the upper absolute band
    | Name: upper\_absolute\_band\_\ **value**\ \_over\_\ **target.name**

    :param value: Constant for band displacement
    :type value: float
    :param target: Data column for the center of the band
    :type target: str
    """
    def return_function(data):
        column_name = f"upper_absolute_band_{value}_over_{target.name}"
        if column_name not in data.columns:
            data[column_name] = value + target
        return data[column_name].copy()

    return return_function


def lower_absolute_band(value, target):
    """
    | Calculates the lower absolute band
    | Name: lower\_absolute\_band\_\ **value**\ \_over\_\ **target.name**

    :param value: Constant for band displacement
    :type value: float
    :param target: Data column for the center of the band
    :type target: str
    """
    def return_function(data):
        column_name = f"lower_absolute_band_{value}_under_{target.name}"
        if column_name not in data.columns:
            data[column_name] = -value + target
        return data[column_name].copy()

    return return_function


def upper_bollinger_band(mean_days, std_days, c):
    """
    | Calculates the upper bollinger band
    | Name: upper\_bollinger\_band\_\ **mean_days**\ \/\ **std_days**

    :param mean_days: Window size for the band center
    :type mean_days: int
    :param std_days: Window size for the band displacement
    :type std_days: str
    :param c: Multiplier constant
    :type c: float
    """
    def return_function(data):
        column_name = f"upper_bollinger_band_{mean_days}/{std_days}"
        if column_name not in data.columns:
            std_dev = rolling_std(std_days)(data)
            ma = moving_average(mean_days)(data)
            data[column_name] = ma + c * std_dev
        return data[column_name].copy()

    return return_function


def lower_bollinger_band(mean_days, std_days, c):
    """
    | Calculates the lower bollinger band
    | Name: lower\_bollinger\_band\_\ **mean_days**\ \/\ **std_days**

    :param mean_days: Window size for the band center
    :type mean_days: int
    :param std_days: Window size for the band displacement
    :type std_days: str
    :param c: Multiplier constant
    :type c: float
    """
    def return_function(data):
        column_name = f"lower_bollinger_band_{mean_days}/{std_days}"
        if column_name not in data.columns:
            std_dev = rolling_std(std_days)(data)
            ma = moving_average(mean_days)(data)
            data[column_name] = ma - c * std_dev
        return data[column_name].copy()

    return return_function


def upper_volatility_band(c, dev_target, band_target, center_target):
    """
    | Calculates the upper volatility band
    | Name: upper\_volatility\_band\_\ **c**\ \_times\_\ **band_target.name**\ &\ **dev_target.name**\ \_over\_\ **center_target.name**

    :param c: Multiplier constant
    :type c: float
    :param dev_target: Used for band displacement. Can be a constant or a function
    :type dev_target: function or float
    :param band_target: Used for band displacement. Can be a constant or a function
    :type band_target: function or float
    :param center_target: Data column for the band center
    :type center_target: str
    """
    def return_function(data):
        if hasattr(band_target, "name") & hasattr(dev_target, "name"):
            column_name = f"upper_volatility_band_{c}_times_{band_target.name}&{dev_target.name}_over_{center_target.name}"
        elif hasattr(band_target, "name"):
            column_name = f"upper_volatility_band_{c}_times_{band_target.name}&{dev_target}_over_{center_target.name}"
        else:
            column_name = f"upper_volatility_band_{c}_times_{band_target}&{dev_target}_over_{center_target.name}"
        if column_name not in data.columns:
            data[column_name] = center_target + c * dev_target * band_target
        return data[column_name].copy()

    return return_function


def lower_volatility_band(c, dev_target, band_target, center_target):
    """
    | Calculates the lower volatility band
    | Name: lower\_volatility\_band\_\ **c**\ \_times\_\ **band_target.name**\ &\ **dev_target.name**\ \_over\_\ **center_target.name**

    :param c: Multiplier constant
    :type c: float
    :param dev_target: Used for band displacement. Can be a constant or a function
    :type dev_target: function or float
    :param band_target: Used for band displacement. Can be a constant or a function
    :type band_target: function or float
    :param center_target: Data column for the band center
    :type center_target: str
    """
    def return_function(data):
        if hasattr(band_target, "name") & hasattr(dev_target, "name"):
            column_name = f"lower_volatility_band_{c}_times_{band_target.name}&{dev_target.name}_under_{center_target.name}"
        elif hasattr(band_target, "name"):
            column_name = f"lower_volatility_band_{c}_times_{band_target.name}&{dev_target}_under_{center_target.name}"
        else:
            column_name = f"lower_volatility_band_{c}_times_{band_target}&{dev_target}_under_{center_target.name}"
        if column_name not in data.columns:
            data[column_name] = center_target - c * dev_target * band_target
        return data[column_name].copy()

    return return_function


def momentum(days=1, target="close"):
    """
    | Calculates the momentum
    | Name: momentum\_\ **days**\ \_of\_\ **target**

    :param days: Window size, defaults to 1
    :type days: int, optional
    :param target: Data column to use, defaults to "close"
    :type target: str, optional
    """
    def return_function(data):
        column_name = f"momentum_{days}_of_{target}"
        if column_name not in data.columns:
            data[column_name] = data[target].diff(periods=days)
        return data[column_name].copy()

    return return_function


def momentum_percentage(days=1, target="close"):
    """
    | Calculates the momentum as a percentage
    | Name: momentum\_percentage\_\ **days**\ \_of\_\ **target**

    :param days: Window size, defaults to 1
    :type days: int, optional
    :param target: Data colum to use, defaults to "close"
    :type target: str, optional
    """
    def return_function(data):
        column_name = f"momentum_percentage_{days}_of_{target}"
        if column_name not in data.columns:
            data[column_name] = data[target].diff(periods=days)
            data.loc[days:, column_name] = data[column_name][days:].divide(
                data[target][: -1 * days].values
            )
        return data[column_name].copy()

    return return_function


def MACD_line(slow_trend=26, fast_trend=12):
    """
    | Calculates the MACD line
    | Name: MACD\_line\_\ **slow_trend**\ \/\ **fast_trend**

    :param slow_trend: Slow window size, defaults to 26
    :type slow_trend: int, optional
    :param fast_trend: Fast window size, defaults to 12
    :type fast_trend: int, optional
    """
    def return_function(data):
        column_name = f"MACD_line_{slow_trend}/{fast_trend}"
        if column_name not in data.columns:
            slow_trend_data = exponential_smoothing(days_to_constant(slow_trend))(data)
            fast_trend_data = exponential_smoothing(days_to_constant(fast_trend))(data)
            data[column_name] = fast_trend_data - slow_trend_data
        return data[column_name].copy()

    return return_function


def MACD_signal(signal, slow_trend=26, fast_trend=12):
    """
    | Calculates the MACD signal
    | Name: MACD\_signal\_\ **slow_trend**\ \/\ **fast_trend**\ \/\ **signal**

    :param signal: Exponential smoothing window size of the MACD line
    :type signal: int
    :param slow_trend: Slow window size of MACD line, defaults to 26
    :type slow_trend: int, optional
    :param fast_trend: Fast windown size of MACD line, defaults to 12
    :type fast_trend: int, optional
    """
    def return_function(data):
        column_name = f"MACD_signal_{slow_trend}/{fast_trend}/{signal}"
        if column_name not in data.columns:
            macd_line = MACD_line(slow_trend, fast_trend)(data)
            data[column_name] = exponential_smoothing(
                days_to_constant(signal), target=macd_line.name
            )(data)
        return data[column_name].copy()

    return return_function


def MACD_histogram(signal, slow_trend=26, fast_trend=12):
    """
    | Calculates the MACD histogram
    | Name: MACD\_histogram\_\ **slow_trend**\ \/\ **fast_trend**\ \/\ **signal**

    :param signal: Exponential smoothing window size of the MACD line
    :type signal: int
    :param slow_trend: Slow window size of MACD line, defaults to 26
    :type slow_trend: int, optional
    :param fast_trend: Fast windown size of MACD line, defaults to 12
    :type fast_trend: int, optional
    """
    def return_function(data):
        column_name = f"MACD_histogram_{slow_trend}/{fast_trend}/{signal}"
        if column_name not in data.columns:
            macd_line = MACD_line(slow_trend, fast_trend)(data)
            macd_signal = MACD_signal(signal, slow_trend, fast_trend)(data)
            data[column_name] = macd_line - macd_signal
        return data[column_name].copy()

    return return_function


def divergence_index(slow_trend, fast_trend):
    """
    | Calculates the divergence index
    | Name: divergence\_index\_\ **slow_trend**\ \/\ **fast_trend**

    :param slow_trend: Slow window size
    :type slow_trend: int
    :param fast_trend: Fast window size
    :type fast_trend: int
    """
    def return_function(data):
        column_name = f"divergence_index_{slow_trend}/{fast_trend}"
        if column_name not in data.columns:
            fast_trend_data = moving_average(fast_trend)(data)
            slow_trend_data = moving_average(slow_trend)(data)
            momentum_data = momentum()(data)
            data[column_name] = (fast_trend_data - slow_trend_data) / rolling_std(
                slow_trend, momentum_data.name
            )(data).pow(2)
        return data[column_name].copy()

    return return_function


def average_true_range(days=10, target="close"):
    """
    | Calculates the average true range
    | Name: average\_true\_range\_\ **days**\ \_of\_\ **target**

    :param days: Window size, defaults to 10
    :type days: int, optional
    :param target: Data column to use, defaults to "close"
    :type target: str, optional
    """
    def return_function(data):
        column_name = f"average_true_range_{days}_of_{target}"
        high_col_name = f"high{target[5:]}"
        low_col_name = f"low{target[5:]}"
        if column_name not in data.columns:
            true_range = data[high_col_name] - data[low_col_name]
            data[column_name] = true_range.rolling(days, min_periods=1).mean()
        return data[column_name].copy()

    return return_function


def stochastic_percentageK(days):
    """
    | Calculates the stochastic K
    | Name: stochastic\_percentageK\_\ **days**

    :param days: Window size
    :type days: int
    """
    def f(x):
        lowest = x["low"].nsmallest(1)
        highest = x["high"].max()
        return 100 * (x["close"].iloc[-1] - lowest) / (highest - lowest)

    def return_function(data):
        column_name = f"stochastic_percentageK_{days}"
        if column_name not in data.columns:
            data[column_name] = np.zeros(len(data))
            column_index = data.columns.get_loc(column_name)
            for i in range(len(data)):
                if i >= days - 1:
                    data.iloc[i, column_index] = f(
                        data.iloc[i - days + 1 : i + 1]
                    ).values
        return data[column_name].copy()

    return return_function


def stochastic_percentageD(days, daysK):
    """
    | Calculates the stochastic D
    | Name: stochastic\_percentageD\_\ **days**\ \/\ **daysK**

    :param days: Window size
    :type days: int
    :param daysK: Window size for the stochastic K
    :type daysK: int
    """
    def return_function(data):
        column_name_K = f"stochastic_percentageK_{daysK}"
        column_name = f"stochastic_percentageD_{days}/{daysK}"
        if column_name not in data.columns:
            stochastic_percentageK(daysK)(data)
            data[column_name] = data[column_name_K].rolling(days).mean()
        return data[column_name].copy()

    return return_function


def ADoscillator():
    """
    | Calculates the A/D oscillator
    | Name: A\/D\_oscillator
    """
    def f(x):
        buying_power = x["high"] - x["open"]
        selling_power = x["close"] - x["low"]
        DRF = (buying_power + selling_power) / 2 / (x["high"] - x["low"])
        return DRF

    def return_function(data):
        column_name = "A/D_oscillator"
        if column_name not in data.columns:
            data[column_name] = f(data)
        return data[column_name].copy()

    return return_function


def percentageR(days):
    """
    | Calculates the stochastic R
    | Name: percentageR\_\ **days**
    
    :param days: Window size
    :type days: int
    """
    def f(x):
        highest = x["high"].max()
        lowest = x["low"].min()
        buying_power = highest - x["close"].iloc[-1]
        return buying_power / (highest - lowest)

    def return_function(data):
        column_name = f"percentageR_{days}"
        if column_name not in data.columns:
            data[column_name] = np.zeros(len(data))
            column_index = data.columns.get_loc(column_name)
            for i in range(len(data)):
                if i >= days - 1:
                    data.iloc[i, column_index] = f(data.iloc[i - days + 1 : i + 1])
        return data[column_name].copy()

    return return_function


def RSI(days):
    """
    | Calculates the RSI
    | Name: RSI\_\ **days**

    :param days: Window size
    :type days: int
    """
    def return_function(data):
        column_name = f"RSI_{days}_2"
        if column_name not in data.columns:
            mom_name = momentum()(data).name
            data['positive_momentum_ma'] = 0
            data.loc[data[mom_name] > 0,'positive_momentum_ma'] = data[data[mom_name] > 0][mom_name]
            data['negative_momentum_ma'] = 0
            data.loc[data[mom_name] < 0,'negative_momentum_ma'] = data[data[mom_name] < 0][mom_name]
            data['AU'] = 0
            data['AD'] = 0
            for i in range(1,len(data)): data.loc[data.index[i], 'AU'] = (data.loc[data.index[i-1],'AU']*13 + data.loc[data.index[i],'positive_momentum_ma'])/14 
            for i in range(1,len(data)): data.loc[data.index[i], 'AD'] = (data.loc[data.index[i-1],'AD']*13 + data.loc[data.index[i],'negative_momentum_ma'])/14 
            data[column_name] = 100 - 100/(1 + data['AU']/(-data['AD']))
        return data[column_name].copy()
    return return_function


def special_k():
    """
    | Calculates the special k indicator
    | Name: special_k
    """
    def return_function(data):
        column_name = "special_k"
        if column_name not in data.columns:
            momentum_days = [10, 15, 20, 30, 40, 65, 75, 100, 195, 265, 390, 530]
            moving_average_days = [10, 10, 10, 15, 50, 65, 75, 100, 130, 130, 130, 195]
            weighted_average_multiplier = [1, 2, 3, 4, 1, 2, 3, 4, 1, 2, 3, 4]
            # Creating all the momentums and moving averages in advance
            ma_names = [
                moving_average(days=ma, target=momentum(days=mom)(data).name)(data).name
                for mom, ma in zip(momentum_days, moving_average_days)
            ]

            data[column_name] = 0
            column_name_pos = data.columns.get_loc(column_name)
            for i in range(0, len(data)):
                data.iloc[i, column_name_pos] = (
                    sum(
                        [
                            weight * data.iloc[i, data.columns.get_loc(ma_name)]
                            for ma_name, weight in zip(
                                ma_names, weighted_average_multiplier
                            )
                        ]
                    )
                    / 30
                )

        return data[column_name].copy()

    return return_function



# -*- coding: utf-8 -*-
"""Functions from market data"""

__author__ = "Miguel Martin"
__version__ = "1"

import numpy as np
from pyquanttrade.engine.utils import (
    max_drawdown_ratio,
    max_drawdown_value,
    safe_div,
    safe_min,
    safe_sum,
    safe_mean,
)
from pyquanttrade.engine.test_summary import TestSummary
from pyquanttrade.engine.stats.daily_stats import DailyStats
from pyquanttrade.engine.stats.summary_stats import SystemSummaryStats
from pyquanttrade.engine.stats.trade_stats import TradeSummaryStats
from pyquanttrade.engine.trade import TradeList


class TestResult:
    def __init__(self, data_sim, ticker, capital):
        self.data = DailyStats(data_sim, ticker, capital)
        self.capital = capital

    def add_ticker(self, data_sim, ticker, capital):
        self.data.add_ticker(data_sim, ticker, capital)   

    def update(self, i, trade_list, ticker):
        self.data.update(i,trade_list, ticker)

    def describe(self, baseline=False):
        self.data.generate_all_stats()

        summary_stats = SystemSummaryStats(self.data)
        trade_summary_stats = {ticker:TradeSummaryStats(self.data.trade_list.clone(ticker), 
                                                        self.data.long_ticker[ticker], 
                                                        self.data.short_ticker[ticker], 
                                                        self.data.global_ticker[ticker], 
                                                        self.data.tickers_capital[ticker]) for ticker in self.data.tickers}
        
        trade_summary_stats['all'] = TradeSummaryStats(self.data.trade_list, self.data.long_all, self.data.short_all, self.data.global_all, np.sum(list(self.data.tickers_capital.values())))
        
        trade_summary_table = {t:v.toDataFrame() for (t,v) in  trade_summary_stats.items()}
        return TestSummary(summary_stats.ticker_stats, trade_summary_table)





from pyquanttrade.engine.commands import backtest, backtest_and_visualise
from tests.policy_battery import test_policy_1
from tests.policy_battery import test_policy_2

def test_backtest():
    result,_ = backtest(test_policy_2,"TSLA", "2012-01-01", "2021-01-01")
    summary_result = result.describe(True)
    assert summary_result

def test_backtest_and_visualise():
    result,_ = backtest_and_visualise(test_policy_2,"TSLA", "2012-01-01", "2021-01-01")
    summary_result = result.describe(True)
    assert summary_result

def test_backest_multiple_stocks():
    stock_list = ['XOM','FB']
    result,_ = backtest(test_policy_2, stock_list, "1999-01-01", "2022-01-01")
    summary_result = result.describe(True)
    assert summary_result

def test_backtest_commission_slippage():
    def slippage(num_actions, open_price, close_price):
        return 0.01 * num_actions * (open_price + close_price)

    def commission(num_actions, open_price, close_price):
        return 0.01 * num_actions * (open_price + close_price)

    result,_ = backtest(test_policy_2,"TSLA", "2012-01-01", "2021-01-01", commission=commission, slippage_perc=slippage)
    summary_result = result.describe(True)
    assert summary_result


# -*- coding: utf-8 -*-
"""Functions from market data"""

__author__ = "Miguel Martin"
__version__ = "1"

import pandas as pd
import numpy as np

from pyquanttrade.engine.utils import (
    max_drawdown_ratio,
    max_drawdown_value,
    safe_div,
    safe_min,
    safe_sum,
    safe_mean,
)

class TradeSummaryStats:

    def __init__(self, trade_list, long_stats, short_stats, system_stats, capital):
        self.all_trades = {}
        self.long_trades = {}
        self.short_trades = {}

        # split long win, long lose, short win and short lose trades for further analysis.
        long_win_trades = [
            trade
            for _, trade in trade_list.trades_closed.items()
            if trade.profit >= 0
            if trade.type == "long"
        ]
        long_lose_trades = [
            trade
            for _, trade in trade_list.trades_closed.items()
            if trade.profit < 0
            if trade.type == "long"
        ]

        short_win_trades = [
            trade
            for _, trade in trade_list.trades_closed.items()
            if trade.profit >= 0
            if trade.type == "short"
        ]
        short_lose_trades = [
            trade
            for _, trade in trade_list.trades_closed.items()
            if trade.profit < 0
            if trade.type == "short"
        ]
        # Num of trades:
        len_long_trades = len(long_lose_trades) + len(long_win_trades)
        len_short_trades = len(short_lose_trades) + len(short_win_trades)
        len_all_trades = len_long_trades + len_short_trades
        self.all_trades['len trades'] = len_all_trades
        self.long_trades['len trades'] = len_long_trades
        self.short_trades['len trades'] = len_short_trades

        self.all_trades['Num of winning trades'] = len(long_win_trades) + len(short_win_trades)
        self.all_trades['Num of losing trades'] = len(long_lose_trades) + len(short_lose_trades)
        self.long_trades['Num of winning trades'] = len(long_win_trades)
        self.long_trades['Num of losing trades'] = len(long_lose_trades)
        self.short_trades['Num of winning trades'] = len(short_win_trades)
        self.short_trades['Num of losing trades'] = len(short_lose_trades)

        # Profit of Winners:
        profit_long_win_trades = np.sum([trade.profit for trade in long_win_trades])
        profit_short_win_trades = np.sum([trade.profit for trade in short_win_trades])
        profit_total_win_trades = profit_long_win_trades + profit_short_win_trades

        self.all_trades['profit win trades'] = profit_total_win_trades
        self.long_trades['profit win trades'] = profit_long_win_trades
        self.short_trades['profit win trades'] = profit_short_win_trades

        # Loss of Lossers:
        lose_long_lose_trades = -np.sum([trade.profit for trade in long_lose_trades])
        lose_short_lose_trades = -np.sum([trade.profit for trade in short_lose_trades])
        lose_total_lose_trades = lose_long_lose_trades + lose_short_lose_trades

        self.all_trades['lose loser trades'] = lose_total_lose_trades
        self.long_trades['lose loser trades'] = lose_long_lose_trades
        self.short_trades['lose loser trades'] = lose_short_lose_trades

        # Net profit:
        net_long = profit_long_win_trades - lose_long_lose_trades
        net_short = profit_short_win_trades - lose_short_lose_trades
        net_total = net_long + net_short

        self.all_trades['net profit'] = net_total
        self.long_trades['net profit'] = net_long
        self.short_trades['net profit'] = net_short

        # Initial Capital:
        self.all_trades['init capital'] = capital
        self.long_trades['init capital'] = capital
        self.short_trades['init capital'] = capital

        # Final Capital:

        self.all_trades['final capital'] = capital + net_total
        self.long_trades['final capital'] = capital + net_long
        self.short_trades['final capital'] = capital + net_short

        # Ratio Net profit:
        long_init_capital = capital
        short_init_capital = capital
        total_init_capital = capital

        ratio_net_long = safe_div(net_long, long_init_capital)
        ratio_net_short = safe_div(net_short, short_init_capital)
        ratio_net_total = safe_div(net_total, total_init_capital)

        self.all_trades['ratio net profit'] = 100 * ratio_net_total
        self.long_trades['ratio net profit'] = 100 * ratio_net_long
        self.short_trades['ratio net profit'] = 100 * ratio_net_short

        # Anualized Ratio Net profit:
        sim_years = system_stats.shape[0] / 250
        self.all_trades['anual ratio net profile'] = 100 * ((ratio_net_total + 1) ** (1 / sim_years) - 1)
        self.long_trades['anual ratio net profile'] = 100 * ((ratio_net_long + 1) ** (1 / sim_years) - 1)
        self.short_trades['anual ratio net profile'] = 100 * ((ratio_net_short + 1) ** (1 / sim_years) - 1)

        # Profit Factor:
        long_profit_factor = safe_div(profit_long_win_trades, lose_long_lose_trades)
        short_profit_factor = safe_div(profit_short_win_trades, lose_short_lose_trades)
        total_profi_factor = safe_div(profit_total_win_trades, lose_total_lose_trades)

        self.all_trades['profit factor'] = total_profi_factor
        self.long_trades['profit factor'] = long_profit_factor
        self.short_trades['profit factor'] = short_profit_factor

        # Percent Profitable:

        long_percent_profitable = safe_div(
            len(long_win_trades), len(long_win_trades + long_lose_trades)
        )
        short_percent_profitable = safe_div(
            len(short_win_trades), len(short_win_trades + short_lose_trades)
        )
        total_percent_profitable = safe_div(
            len(long_win_trades + short_win_trades),
            len(
                long_win_trades
                + short_win_trades
                + long_lose_trades
                + short_lose_trades
            ),
        )

        self.all_trades['% profitable'] = total_percent_profitable
        self.long_trades['% profitable'] = long_percent_profitable
        self.short_trades['% profitable'] = short_percent_profitable

        # Payoff Ratio:

        long_avg_win = safe_div(profit_long_win_trades, len(long_win_trades))
        short_avg_win = safe_div(profit_short_win_trades, len(short_win_trades))
        total_avg_win = safe_div(
            profit_total_win_trades, len(long_win_trades + short_win_trades)
        )

        long_avg_lose = safe_div(lose_long_lose_trades, len(long_lose_trades))
        short_avg_lose = safe_div(lose_short_lose_trades, len(short_lose_trades))
        total_avg_lose = safe_div(
            lose_total_lose_trades, len(long_lose_trades + short_lose_trades)
        )

        long_payoff_ratio = safe_div(long_avg_win, long_avg_lose)
        short_payoff_ratio = safe_div(short_avg_win, short_avg_lose)
        total_payoff_ratio = safe_div(total_avg_win, total_avg_lose)

        self.all_trades['payoff ratio'] = total_payoff_ratio
        self.long_trades['payoff ratio'] = long_payoff_ratio
        self.short_trades['payoff ratio'] = short_payoff_ratio

        # Maximum loss:

        long_max_loss = safe_min([trade.profit for trade in long_lose_trades])
        short_max_loss = safe_min([trade.profit for trade in short_lose_trades])
        total_max_loss = np.max([long_max_loss, short_max_loss])

        self.all_trades['max loss'] = total_max_loss
        self.long_trades['max loss'] = long_max_loss
        self.short_trades['max loss'] = short_max_loss

        # Maximum consecutive lossing trades:
        long_max_consecutives_lossing_trades = 0
        short_max_consecutives_lossing_trades = 0
        total_max_consecutives_lossing_trades = 0

        long_current_consecutives_lossing_trades = 0
        short_current_consecutives_lossing_trades = 0
        total_current_consecutives_lossing_trades = 0

        for k, trade in trade_list.trades_closed.items():
            if trade.type == "long" and trade.profit < 0:
                long_current_consecutives_lossing_trades += 1
                total_current_consecutives_lossing_trades += 1

            if trade.type == "short" and trade.profit < 0:
                short_current_consecutives_lossing_trades += 1
                total_current_consecutives_lossing_trades += 1

            if trade.type == "long" and trade.profit >= 0:
                if (
                    long_current_consecutives_lossing_trades
                    > long_max_consecutives_lossing_trades
                ):
                    long_max_consecutives_lossing_trades = (
                        long_current_consecutives_lossing_trades
                    )
                long_current_consecutives_lossing_trades = 0

                if (
                    total_current_consecutives_lossing_trades
                    > total_max_consecutives_lossing_trades
                ):
                    total_max_consecutives_lossing_trades = (
                        total_current_consecutives_lossing_trades
                    )
                total_current_consecutives_lossing_trades = 0

            if trade.type == "short" and trade.profit >= 0:
                if (
                    short_current_consecutives_lossing_trades
                    > short_max_consecutives_lossing_trades
                ):
                    short_max_consecutives_lossing_trades = (
                        short_current_consecutives_lossing_trades
                    )
                short_current_consecutives_lossing_trades = 0

                if (
                    total_current_consecutives_lossing_trades
                    > total_max_consecutives_lossing_trades
                ):
                    total_max_consecutives_lossing_trades = (
                        total_current_consecutives_lossing_trades
                    )
                total_current_consecutives_lossing_trades = 0

        if (
            long_current_consecutives_lossing_trades
            > long_max_consecutives_lossing_trades
        ):
            long_max_consecutives_lossing_trades = (
                long_current_consecutives_lossing_trades
            )

        if (
            short_current_consecutives_lossing_trades
            > short_max_consecutives_lossing_trades
        ):
            short_max_consecutives_lossing_trades = (
                short_current_consecutives_lossing_trades
            )

        if (
            total_current_consecutives_lossing_trades
            > total_max_consecutives_lossing_trades
        ):
            total_max_consecutives_lossing_trades = (
                total_current_consecutives_lossing_trades
            )

        self.all_trades['max consecutives lossing trades'] = total_max_consecutives_lossing_trades
        self.long_trades['max consecutives lossing trades'] = long_max_consecutives_lossing_trades
        self.short_trades['max consecutives lossing trades'] = short_max_consecutives_lossing_trades

        # exposure:

        long_exposure = safe_div(
            long_stats[long_stats["open_trades"] > 0][
                "open_trades"
            ].count(),
            long_stats["open_trades"].count(),
        )

        short_exposure = safe_div(
            short_stats[short_stats["open_trades"] > 0][
                "open_trades"
            ].count(),
            short_stats["open_trades"].count(),
        )

        global_exposure = safe_div(
            system_stats[system_stats["open_trades"] > 0][
                "open_trades"
            ].count(),
            system_stats["open_trades"].count(),
        )

        self.all_trades['exposure'] = global_exposure * 100
        self.long_trades['exposure'] = long_exposure * 100
        self.short_trades['exposure'] = short_exposure * 100

        # holding period:

        long_holding = safe_mean(
            [
                trade.num_days
                for k, trade in trade_list.trades_closed.items()
                if trade.type == "long"
            ]
        )
        short_holding = safe_mean(
            [
                trade.num_days
                for k, trade in trade_list.trades_closed.items()
                if trade.type == "short"
            ]
        )
        global_holding = safe_mean(
            [trade.num_days for k, trade in trade_list.trades_closed.items()]
        )

        self.all_trades['holding'] = global_holding
        self.long_trades['holding'] = long_holding
        self.short_trades['holding'] = short_holding

        # Length of the average winning trade to the average losing trade:

        long_wins_holding = safe_mean(
            [
                trade.num_days
                for k, trade in trade_list.trades_closed.items()
                if trade.type == "long" and trade.profit > 0
            ]
        )
        long_lose_holding = safe_mean(
            [
                trade.num_days
                for k, trade in trade_list.trades_closed.items()
                if trade.type == "long" and trade.profit <= 0
            ]
        )
        long_holding_ratio = safe_div(long_wins_holding, long_lose_holding)

        short_wins_holding = safe_mean(
            [
                trade.num_days
                for k, trade in trade_list.trades_closed.items()
                if trade.type == "short" and trade.profit > 0
            ]
        )
        short_lose_holding = safe_mean(
            [
                trade.num_days
                for k, trade in trade_list.trades_closed.items()
                if trade.type == "short" and trade.profit <= 0
            ]
        )
        short_holding_ratio = safe_div(short_wins_holding, short_lose_holding)

        global_wins_holding = safe_mean(
            [
                trade.num_days
                for k, trade in trade_list.trades_closed.items()
                if trade.profit > 0
            ]
        )
        global_lose_holding = safe_mean(
            [
                trade.num_days
                for k, trade in trade_list.trades_closed.items()
                if trade.profit <= 0
            ]
        )
        global_holding_ratio = safe_div(global_wins_holding, global_lose_holding)

        self.all_trades['holding ratio'] = global_holding_ratio
        self.long_trades['holding ratio'] = long_holding_ratio
        self.short_trades['holding ratio'] = short_holding_ratio

        # profit expectancy.

        l_trades = long_win_trades + long_lose_trades
        s_trades = short_win_trades + short_lose_trades
        g_trades = l_trades + s_trades

        long_expectancy = safe_mean([trade.profit for trade in l_trades])
        short_expectancy = safe_mean([trade.profit for trade in s_trades])
        global_expectancy = safe_mean([trade.profit for trade in g_trades])

        self.all_trades['expectancy'] = global_expectancy
        self.long_trades['expectancy'] = long_expectancy
        self.short_trades['expectancy'] = short_expectancy

        # maxDD

        l_max_drawn_down = max_drawdown_value(
            long_stats["profit_sequence"]
        )
        s_max_drawn_down = max_drawdown_value(
            short_stats["profit_sequence"]
        )
        g_max_drawn_down = max_drawdown_value(system_stats["profit_sequence"])

        self.all_trades['max draw drown']= g_max_drawn_down
        self.long_trades['max draw drown'] = l_max_drawn_down
        self.short_trades['max draw drown'] = s_max_drawn_down

        # Recovery ratio

        self.all_trades['recovery ratio'] = safe_div(net_total, g_max_drawn_down)
        self.long_trades['recovery ratio'] = safe_div(net_long, l_max_drawn_down)
        self.short_trades['recovery ratio'] = safe_div(net_short, s_max_drawn_down)

        self.all_trades['avg_commission_cost'] = safe_mean([trade.commission_cost for _,trade in trade_list.trades_closed.items()])
        self.long_trades['avg_commission_cost'] = safe_mean([trade.commission_cost for _,trade in trade_list.trades_closed.items() if trade.type == 'long'])
        self.short_trades['avg_commission_cost'] = safe_mean([trade.commission_cost for _,trade in trade_list.trades_closed.items() if trade.type == 'short'])

        self.all_trades['avg_slippage_cost'] = safe_mean([trade.slippage_cost for _,trade in trade_list.trades_closed.items()])
        self.long_trades['avg_slippage_cost'] = safe_mean([trade.slippage_cost for _,trade in trade_list.trades_closed.items() if trade.type == 'long'])
        self.short_trades['avg_slippage_cost'] = safe_mean([trade.slippage_cost for _,trade in trade_list.trades_closed.items() if trade.type == 'short'])


    def toDataFrame(self):
        idx = [
            "Num of trades",
            "Num of winning trades",
            "Num of losing trades",
            "Profit of Winners",
            "Loss of Lossers",
            "Net profit",
            "Initial capital",
            "Final Capital",
            "% Net profit",
            "Anualized % Net profit",
            "Profit Factor",
            "Percent Profitable",
            "Payoff Ratio",
            "Maximum loss",
            "Maximum consecutive lossing trades",
            " % exposure",
            "Holding period",
            "lengt of avg win / leng of avg lose",
            "Profit expectancy",
            "MaxDD",
            "Recovery Ratio",
            "Average commission cost",
            "Average slippage cost"
        ]

        data = {
            "All trades": list(self.all_trades.values()),
            "Long trades": list(self.long_trades.values()),
            "Short trades": list(self.short_trades.values()),
        }

        all_trade_statistics = pd.DataFrame(data, index=idx)

        return all_trade_statistics.round(2)




import setuptools
from setuptools import find_packages
setuptools.setup(
    name="pyquanttrade",  
    version="0.0.15",
    author="Miguel Martin, Marcos Jimenez",
    description="Library for backtesting algorithmic trading strategies using Quandl data",
    
    packages=['pyquanttrade','pyquanttrade.features', 'pyquanttrade.engine', 'pyquanttrade.engine.stats'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[
        "pandas",
        "datetime",
        "IPython",
        "altair",
        "quandl",
        "python-decouple",
        'scipy',
        'plotly',
        'tqdm'
    ],
)



# -*- coding: utf-8 -*-
"""Class for get Market data"""

__author__ = "Miguel Martin"
__version__ = "1"

import pandas as pd
import numpy as np
import altair as alt
import scipy.stats as stats
import scipy.spatial.distance as distance
import time, quandl
from decouple import config
class marketData(object):

    @staticmethod
    def get_data(ticker, init_date, end_date):
        quandl.ApiConfig.api_key=config("QUANDL_API_KEY")
        table_name = 'SHARADAR/SEP'
        filter_date = { 'gte': init_date, 'lte':end_date}
        source = quandl.get_table(table_name,  date = filter_date, ticker=ticker, paginate=True)
        source = pd.DataFrame(source).drop_duplicates("date")
        source["date"] = pd.to_datetime(source["date"])
        source = source.sort_values(by="date")
        source = source.set_index("date")
        return source

    @staticmethod
    def get_data_n_tries(ticker, init_date, end_date, n_tries = 5):
        tried = 0
        completed = False
        while tried<n_tries and completed == False:
            try:
                data = marketData.get_data(ticker, init_date, end_date)
                completed = True
            except:
                time.sleep(2000)
                tried += 1
        assert tried<n_tries, 'The data could not be retrieved after ' + str(n_tries) + ' tries'
        return data

    @staticmethod
    def get_data_lt(ticker, init_date):
        quandl.ApiConfig.api_key=config("QUANDL_API_KEY")
        table_name = 'SHARADAR/SEP'
        filter_date = {'gte':init_date}
        source = quandl.get_table(table_name,  date = filter_date, ticker=ticker, paginate=True)
        source = pd.DataFrame(source).drop_duplicates("date")
        source["date"] = pd.to_datetime(source["date"])
        source = source.sort_values(by="date")
        source = source.set_index("date")
        return source

    @staticmethod
    def get_data_and_visualize(
        ticker, init_date, end_date, moving_averages=[], ribbons=False
    ):
        # Get the data from ElasticSearch

        source = marketData.get_data(ticker, init_date, end_date)
        # visualize
        open_close_color = alt.condition(
            "datum.open < datum.close", alt.value("#06982d"), alt.value("#ae1325")
        )
        rule = (
            alt.Chart(source.reset_index())
            .mark_rule()
            .encode(
                alt.X(
                    "yearmonthdate(date):T",
                ),
                alt.Y("low", scale=alt.Scale(zero=False), axis=alt.Axis(title="Price")),
                alt.Y2("high"),
                color=open_close_color,
            )
            .properties(width=1000, height=400)
        )
        bar = (
            alt.Chart(source.reset_index())
            .mark_bar()
            .encode(
                x="yearmonthdate(date):T", y="open", y2="close", color=open_close_color
            )
        )
        vis = rule + bar
        colors = [
            "blue",
            "grey",
            "orange",
            "brown",
            "yellow",
            "pink",
            "black",
            "violet",
            "red",
            "cyan",
        ]
        for idx, m in enumerate(moving_averages):
            field = "mean_" + str(m)
            source[field] = source["close"].rolling(m, min_periods=1).mean()
            color_line = colors[idx]
            mean = (
                alt.Chart(source.reset_index())
                .mark_line()
                .encode(
                    alt.X("yearmonthdate(date):T"),
                    alt.Y(field),
                    color=alt.value(color_line),
                )
            )
            vis = vis + mean

        source["mean_30"] = source["close"].rolling(10, min_periods=1).mean()
        source["std_dev"] = source["close"].rolling(10, min_periods=1).std()
        source["std_dev_up"] = source["mean_30"] + 2 * source["std_dev"]
        source["std_dev_down"] = source["mean_30"] - 2 * source["std_dev"]

        area = (
            alt.Chart(source.reset_index())
            .mark_area(opacity=0.3)
            .encode(x="yearmonthdate(date):T", y=alt.Y("std_dev_up"), y2="std_dev_down")
        )

        vis = vis + area
        if ribbons:
            rolling_means = {}
            for i in np.linspace(10, 100, 10):
                X = source["close"].rolling(window=int(i), min_periods=1).mean()
                rolling_means[i] = X

            rolling_means = pd.DataFrame(rolling_means).dropna()

            scores = pd.Series(index=source.index)

            for date in rolling_means.index:
                mavg_values = rolling_means.loc[date]
                ranking = stats.rankdata(mavg_values.values)
                d = distance.hamming(ranking, range(1, 11))
                scores[date] = d

            # Normalize the  score
            source["ribbon"] = scores.rolling(50, min_periods=1).mean()
            source["ribbon_velocity"] = source["ribbon"].diff()
            ribbon1 = (
                alt.Chart(source.reset_index())
                .mark_line()
                .encode(
                    alt.X("yearmonthdate(date):T"),
                    alt.Y("ribbon"),
                    color=alt.value("blue"),
                )
            )

            ribbon2 = (
                alt.Chart(source.reset_index())
                .mark_line()
                .encode(
                    alt.X("yearmonthdate(date):T"),
                    alt.Y("ribbon_velocity"),
                    color=alt.value("red"),
                )
            )
            ribbon = ribbon1 + ribbon2
            vis = alt.layer(vis, ribbon).resolve_scale(y="independent")

        return vis, source






from pyquanttrade.features.functions import (
    exponential_smoothing,
    moving_average,
    upper_bollinger_band,
    lower_bollinger_band,
    get_column,
    days_to_constant
)
from pyquanttrade.policy import Policy

from pyquanttrade.features.indicators import cross_of_values


class test_policy_1(Policy):

    name = "bollinger cross"
    version = "1.0"

    # stop loss parameters:
    long_stop_loss = 100
    short_stop_loss = 100
    long_stop_loss_trailling = 100
    short_stop_loss_trailling = 100

    def description(self):
        return "Primera estrategia de prueba con reversión a la media"

    # def sell_short_when(self):
    # func1 = get_column('close')
    # func2 = lower_bollinger_band(20,20,2)
    # return cross_of_values(func2,func1)
    @classmethod
    def buy_long_when(self):
        lower_bollinger = lower_bollinger_band(50, 50, 2)
        return cross_of_values(lower_bollinger, get_column("close"))

    # def close_short_when(self):
    # func1 = get_column('close')
    # func2 = moving_average(20)
    # return cross_of_values(func1,func2)
    @classmethod
    def close_long_when(self):
        upper_bollinger = upper_bollinger_band(50, 50, 2)
        option1 = cross_of_values(get_column("close"), upper_bollinger)
        # option2 = downwards_turn(moving_average(30))
        # return apply_any([option1, option2])
        return option1

class test_policy_2(Policy):

    name = "long/short EMA crossover 10/40"
    version = "1.0"

    # stop loss parameters:
    long_stop_loss = 0.5
    short_stop_loss = 100
    long_stop_loss_trailling = 100
    short_stop_loss_trailling = 100

    def description(self):
        return "Estrategia de tipo moving average crossover. Inversión en largo cuando la SMA_10 supera SMA_40. Inversión en corto cuando la SMA_10 cae por debajo de SMA_40. Las posiciones corto/largo se cierran cuando se abren las contrarias."

    @classmethod
    def buy_long_when(self):
        func1 = exponential_smoothing(days_to_constant(50))
        func2 = moving_average(40)
        return cross_of_values(func1, func2)

    #def sell_short_when(self):
    #    func1 = moving_average(10)
    #    func2 = moving_average(40)
    #    return cross_of_values(func2, func1)
    @classmethod
    def close_long_when(self):
        func1 = moving_average(10)
        func2 = moving_average(40)
        return cross_of_values(func2, func1)

    #def close_short_when(self):
    #    func1 = moving_average(10)
    #    func2 = moving_average(40)
    #    return cross_of_values(func1, func2)